#!/usr/bin/perl 
package LLUtil;

# static char sccsid[] = "@(#)15 1.17 src/ll/config_editor/cgi-bin/LLUtil.pm, ll.config, ll_rrnep2, rrnep2s015a 3/27/13 17:32:41\n";
use DBI;
use strict;
# if AIX - make sure we include perl 5.8.2 in INC path.
#       Needed to find perl dependencies shipped in deps tarball.
if ($^O =~ /^aix/i) {
        use lib "/usr/opt/perl5/lib/5.8.2/aix-thread-multi";
        use lib "/usr/opt/perl5/lib/5.8.2";
        use lib "/usr/opt/perl5/lib/site_perl/5.8.2/aix-thread-multi";
        use lib "/usr/opt/perl5/lib/site_perl/5.8.2";
}
use CGI;
use CGI::Session;
use HTML::Template;
use Math::BigInt;

our @EXPORT = qw (get_db_info get_dbh debug_log unescape rsa_encrypt rsa_decrypt get_only_one error_and_exit delete_by_id check_data format_keywords get_value_by_id
                  get_default_keyword_str get_default_id_by_primary_key get_default_items get_default_json_str get_default_cluster_id get_default_id
                  get_fsID_by_nodeID get_default_json_str_of_machine_env get_default_json_str_of_preempt get_admin);
use Exporter;
our @ISA = qw(Exporter);

use constant TMPL_DIR => "/var/www/ll/html";
use constant OUTPUT_HEADER => "Content-type: text/html\n\n";


use constant ALL_TABLES => { tab_panel_classes => [
						   # normal tables by classID
						   ["TLLS_CFGClass"],
						   # editable tables by classID
						   ["TLLS_CFGClassDefaultResources","TLLS_CFGClassMaxResources","TLLS_CFGClassMaxNodeResources",
						    "TLLS_CFGClassDefaultNodeResources","TLLS_CFGClassLimits","TLLS_CFGClassUser"],
						   # multiple select tables by classID
						   ["TLLS_CFGClassGroups","TLLS_CFGClassUsers","TLLS_CFGClassBG"]
						  ],
			     tab_panel_classes_type => ["normal","editable","multiselect"],
			     
			     tab_panel_cluster_env => [
						       # normal tables by clusterID
						       ["TLL_CFGCluster","TLLR_CFGRMCluster","TLLR_CFGExternalScheduler"],
						       # editable tables by clusterID
						       ["TLLR_CFGRMClusterFloatingResources","TLLR_CFGVipServerList"]
						      ],
			     tab_panel_cluster_env_type => ["normal","editable"],
			     
			     tab_panel_daemons => [
						   # normal tables in tab_panel_daemons
						   ["TLLR_CFGDaemons"]
						  ],
			     tab_panel_daemons_type => ["normal"],
			     
			     tab_panel_cm => [
					      # normal tables in tab_panel_cm
					      ["TLLS_CFGCentralManager"]
					     ],
			     tab_panel_cm_type => ["normal"],
			     
			     tab_panel_resmgr => [
						  # normal tables in tab_panel_resmgr
						  ["TLLR_CFGResourceMgr"]
						 ],
			     tab_panel_resmgr_type => ["normal"],
			     
			     tab_panel_regmgr => [
						  # normal tables in tab_panel_regmgr
						  ["TLLR_CFGRegionMgr"]
						 ],
			     tab_panel_regmgr_type => ["normal"],
			     
			     tab_panel_master => [
						  # normal tables in tab_panel_master
						  ["TLLR_CFGMaster"]
						 ],
			     tab_panel_master_type => ["normal"],
			     
			     tab_panel_startd => [
						  # normal tables in tab_panel_startd
						  ["TLLR_CFGStartd"]
						 ],
			     tab_panel_startd_type => ["normal"],
			     
			     tab_panel_starter => [
						   # normal tables in tab_panel_starter
						   ["TLLR_CFGStarter"]
						  ],
			     tab_panel_starter_type => ["normal"],
			     
			     tab_panel_schedd => [
						  # normal tables in tab_panel_schedd
						  ["TLLR_CFGSchedd"]
						 ],
			     tab_panel_schedd_type => ["normal"],
			     
			     tab_panel_kbdd => [
						# normal tables in tab_panel_kbdd
						["TLLR_CFGKbdd"]
					       ],
			     tab_panel_kbdd_type => ["normal"],
			     
			     tab_panel_groups => [
						  # normal tables by groupID
						  ["TLLS_CFGGroup"],
						  # multiple select tables by groupID
						  ["TLLS_CFGGroupUsers"]
						 ],
			     tab_panel_groups_type => ["normal","multiselect"],
			     
			     tab_panel_machine_env => [
						       # normal tables by nodeID
						       ["TLL_CFGCommon","TLLR_CFGAcct","TLLR_CFGWLM","TLLR_CFGSwitch",
							"TLLR_CFGCkpt","TLLR_CFGSecurity","TLL_CFGRunPolicy"],
						       # editable tables by nodeID
						       ["TLLR_CFGAcctFlags","TLLR_CFGWLMUsage"],
						       # normal tables by fsID
						       ["TLLR_CFGFileSysMon"],
						       # editable tables by fsID
						       ["TLLR_CFGFileSysMonThresholds"]
						      ],
			     tab_panel_machine_env_type => ["normal","editable","normal","editable"],

			     tab_panel_machine_groups => [
							  # normal tables by mgID
							  ["TLL_MachineGroup","TLLR_CFGMachineGroup"],
							  # editable tables by mgID
							  ["TLLR_CFGMachineGroupMachineList","TLLR_CFGMachineGroupClass","TLLR_CFGMachineGroupResources"],
							  # multiple select tables by mgID
							  ["TLLR_CFGMachineGroupNameServer"]
							 ],
			     tab_panel_machine_groups_type => ["normal","editable","multiselect"],

			     tab_panel_machines => [
						    # normal tables by nodeID
						    ["TLL_Nodelist","TLLR_CFGMachine"],
						    # editable tables by nodeID
						    ["TLLR_CFGMachineClass","TLLR_CFGMachineResources"],
						    # multiple select tables by nodeID
						    ["TLLR_CFGMachineNameServer"]
						   ],
			     tab_panel_machines_type => ["normal","editable","multiselect"],
			     
			     tab_panel_multicluster => [
							# normal tables by clusterID
							["TLL_Cluster","TLLS_CFGMCluster","TLLR_CFGMCluster"],
							# editable tables by clusterID
							["TLLR_CFGMClusterClasses","TLLR_CFGMClusterGroups","TLLR_CFGMClusterUsers",
							 "TLLR_CFGMClusterInboundHosts","TLLR_CFGMClusterOutboundHosts"]
						       ],
			     tab_panel_multicluster_type => ["normal","editable"],
			     
			     tab_panel_region => [
						  # normal tables by regionID
						  ["TLLR_CFGRegion"]
						 ],
			     tab_panel_region_type => ["normal"],

			     tab_panel_users => [
						 # normal tables by userID
						 ["TLLS_CFGUser"]
						],
			     tab_panel_users_type => ["normal"],
			     
			     tab_panel_variables => [
						     # editable tables by clusterID
						     ["TLLR_CFGVars"]
						    ],
			     tab_panel_variables_type => ["editable"],
			     
			     tab_panel_scheduler => [
						     # normal tables by clusterID in tab_panel_scheduler
						     ["TLLS_CFGScheduler"],
						     # editable tables by clusterID in tab_panel_scheduler
						     ["TLLS_CFGScheduleByResources"]
						    ],
			     tab_panel_scheduler_type => ["normal","editable"],
			     
			     tab_panel_preempt => [
						   # normal tables by nodeID in tab_panel_preempt
						   ["TLLS_CFGPreempt"]
						  ],
			     tab_panel_preempt_type => ["normal"],
			     
			     tab_panel_reservation => [
						       # normal tables by clusterID in tab_panel_reservation
						       ["TLLS_CFGReservation"]
						      ],
			     tab_panel_reservation_type => ["normal"],
			     
			     tab_panel_bg => [
					      # normal tables by clusterID in tab_panel_bg
					      ["TLLS_CFGBG"]
					     ],
			     tab_panel_bg_type => ["normal"],
			     
			     tab_panel_fairshare => [
						     # normal tables by clusterID in tab_panel_fairshare
						     ["TLLS_CFGFairShare"]
						    ],
			     tab_panel_fairshare_type => ["normal"],
			     
			     tab_panel_dstg => [
						# normal tables by clusterID in tab_panel_dstg
						["TLLS_CFGDstg"]
					       ],
			     tab_panel_dstg_type => ["normal"]
			     
			   };


#########################################################
#                                                       #
#               1. General Functions                    #
#                                                       #
#########################################################
sub get_db_info
{
	my $db_host = undef;
	my $db_db = undef;
	my $db_user = undef;
	my $db_pass = undef;
	my $db_conn = undef;
	my $db2inst = undef;

	my $cgi = CGI->new;
	my $sid = $cgi->cookie("CGISESSID") || undef;
	if(defined($sid))
	{
		my $session = new CGI::Session(undef, $sid, {Directory=>'/tmp'});
		$db_user = rsa_decrypt($session->param("db_user"));
		$db_pass = rsa_decrypt($session->param("db_pass"));
		$db_host = rsa_decrypt($session->param("db_host"));
		$db_db = rsa_decrypt($session->param("db_db"));
		$db_conn = rsa_decrypt($session->param("db_conn"));
		$db2inst = rsa_decrypt($session->param("db2inst"));
		
		$session->expire("30m");
#		debug_log("read out:".$session->param("db_user")."-".$session->param("db_pass")."-".$session->param("db_host")."-".$session->param("db_db")."-".$session->param("db_db"));
		debug_log("read out:".$db_user."-".$db_pass."-".$db_host."-".$db_db."-".$db_conn."-".$db2inst);
	}
	
	if( (!defined($db_user)) || (!defined($db_pass)) || (!defined($db_host)) || (!defined($db_db)) || (!defined($db_conn)))
	{
		my $return_tmpl = $_[0];
		if($return_tmpl == 1)
		{
			use constant TMPL_FILE => TMPL_DIR."/llconfig_login.tmpl";
			my $tmpl = new HTML::Template(filename => TMPL_FILE);
			print OUTPUT_HEADER.$tmpl->output;
			exit;
		}
		else
		{
			print OUTPUT_HEADER."Login";
			exit;
		}
	}
	
	($db_host,$db_user,$db_pass,$db_db,$db_conn,$db2inst);
}

sub get_dbh
{
	my $return_tmpl = $_[0];

	my ($db_host, $db_user, $db_pass, $db_db, $db_conn, $db2inst) = get_db_info($return_tmpl);
	if ($db_conn =~ /^DB2/ && defined($db2inst))
	{
		$ENV{'DB2INSTANCE'} = $db2inst; 
	}
	my $dbh = DBI->connect("DBI:".$db_conn,$db_user,$db_pass);
	$dbh;
}

sub debug_log
{
# 	open LOG, ">>/tmp/forms_based_editor.log";
# 	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime();
# 	print LOG $hour.":".$min.":".$sec.": ".$_[0]."\n";
# 	print LOG "\n";
# 	close LOG;
}

sub unescape
{
    $_[0] =~ s/%20/ /g;
    $_[0];
}

sub rsa_encrypt
{
	my %RSA_CORE = (n=>2773,e=>63,d=>847);
	my $N=new Math::BigInt($RSA_CORE{n});
	my $E=new Math::BigInt($RSA_CORE{e});
	my $D=new Math::BigInt($RSA_CORE{d});

	my $src = shift @_;
	my $result;
	for(my $i=0;$i < length($src);$i++)
	{
		my $c=ord(substr($src,$i,1));
		my $M=Math::BigInt->new($c);
		my $C=$M->copy();
		$C->bmodpow($D,$N);
		$c=sprintf("%03X",$C);
		$result.=$c;
	}
	return $result;
}

sub rsa_decrypt
{
	my %RSA_CORE = (n=>2773,e=>63,d=>847);
	my $N=new Math::BigInt($RSA_CORE{n});
	my $E=new Math::BigInt($RSA_CORE{e});
	my $D=new Math::BigInt($RSA_CORE{d});

	my $src = shift @_;
	my $result;
	for(my $i=0;$i < length($src);$i+=3)
	{
		my $c=substr($src,$i,3);
		$c=hex($c);
		my $M=Math::BigInt->new($c);
		my $C=$M->copy();
		$C->bmodpow($E,$N);
		$c=chr($C);
		$result.=$c;
	}
	debug_log("decrypt return = ".$result);
	return $result;
} 

# Different syntax for DB2, MySQL
# get_only_one($db_conn)
sub get_only_one
{
	my $result;
	if ($_[0] =~ /^DB2/)
	{
		$result = " fetch first 1 rows only"; 
	}
	#  MySQL
	else
	{
		$result = " limit 1";
	}
	return $result;

}
#########################################################
#                                                       #
#               2. Functions for default values         #
#                                                       #
#########################################################

# error_and_exit($dbh->errstr)
sub error_and_exit
{
	print OUTPUT_HEADER."Error:".$_[0];
	exit 1;
}

# delete_by_id($dbh,"TLL_Nodelist","nodeID",$return_id)
sub delete_by_id
{
	my $dbh = $_[0];
	my $table = $_[1];
	my $id1 = $_[2],
	my $primary_key_value = $_[3];

	my $sql_str = "delete from ".$table." where ".$id1."=".$primary_key_value;
#	debug_log($sql_str);
	if(!defined($dbh->do($sql_str)))
	{
		error_and_exit($dbh->errstr);
	}
}

# get_value_by_id($dbh,"name","classID","TLLS_CFGClass",$columns->[0])
sub get_value_by_id
{
	my $dbh = $_[0];
	my $name1 = $_[1];
	my $id1 = $_[2],
	my $table = $_[3];
	my $primary_key_value = $_[4];

	my $sql_str = "select ".$name1." from ".$table." where ".$id1."=".$primary_key_value;

	my $ret = undef;
	
	my $sth=$dbh->prepare($sql_str);
	my($x);
	if(!defined($sth->execute()))
	{
		error_and_exit($dbh->errstr);
	}
	else
	{
		$sth->bind_columns(undef,\$x);
		if($sth->fetch())
		{
			$ret = $x;
		}
	}
	$ret;
}

# format_keywords($dbh,\@columns, $table)
# Note: this function does not format string for preempt_class and start_class.
sub format_keywords
{
	my $dbh = $_[0];
	my $columns = $_[1];
	my $table = $_[2];

	my $rs = "";
	
	my $n = scalar(@$columns);
	
	if($table eq "TLLS_CFGClassDefaultResources" || $table eq "TLLS_CFGClassMaxResources" || $table eq "TLLS_CFGClassMaxNodeResources" ||
	   $table eq "TLLS_CFGClassDefaultNodeResources" || $table eq "TLLR_CFGRMClusterFloatingResources" ||
	   $table eq "TLLR_CFGMachineGroupClass" || $table eq "TLLR_CFGMachineGroupResources" ||
	   $table eq "TLLR_CFGMachineClass" || $table eq "TLLR_CFGMachineResources" ||
	   $table eq "TLLR_CFGMClusterInboundHosts" || $table eq "TLLR_CFGMClusterOutboundHosts" ||
	   # no keyword for "TLLR_CFGMClusterClasses","TLLR_CFGMClusterGroups","TLLR_CFGMClusterUsers" in TLL_CFGKeywords ?
	   $table eq "TLLR_CFGMClusterClasses" || $table eq "TLLR_CFGMClusterGroups" || $table eq "TLLR_CFGMClusterUsers")
	{
		$rs .= $columns->[1]."(".$columns->[2].") ";
	}
	elsif($table eq "TLLS_CFGClassLimits" || $table eq "TLLR_CFGFileSysMonThresholds")
	{
		# only return the value pair
		$rs .= $columns->[2].",".$columns->[3]." ";
	}
	elsif($table eq "TLLS_CFGClassUser")
	{
		my $class_name = get_value_by_id($dbh,"name","classID","TLLS_CFGClass",$columns->[0]);
		$rs .= $class_name.":{type=class ".$columns->[1].":{ type=user maxidle=".$columns->[3]." maxjobs=".$columns->[4]." maxqueued=".$columns->[5]." max_total_tasks=".$columns->[2]."} }";
	}
	elsif($table eq "TLLS_CFGPreemptClass")
	{
		my $outgoing_str = "";
		my @outgoing_classes = ();
		if($columns->[0] != -1)
		{
			@outgoing_classes = @{get_outgoing_classes($dbh,$columns->[0])};
		}
		else
		{
			@outgoing_classes = split(' ',$columns->[5]);
		}
		foreach my $class (@outgoing_classes)
		{
			if(length($class) > 0) { $outgoing_str .= $class." "; }			
		}
		$outgoing_str =~ s/ $//g;
		$rs .= "PREEMPT_CLASS[".$columns->[2]."]=".$columns->[3].":".$columns->[4]." {".$outgoing_str."}"." ";
	}
	elsif($table eq "TLLS_CFGPreemptStartClass")
	{
		my $running_str = "";
		my @running_tasks = ();
		if($columns->[0] != -1)
		{
			@running_tasks = @{get_running_tasks($dbh,$columns->[0])};
			foreach my $task (@running_tasks)
			{
				$running_str .= $task." && ";
			}
			$running_str =~ s/ && $//g;
		}
		else
		{
			# my @columns = (-1,$values{"clusterID"},$values{"incoming_class"},$values{"running_class"},$values{"number_of_tasks"});
			$running_str = "(".$columns->[3]."<".$columns->[4].")";
		}
		$rs .= "START_CLASS[".$columns->[2]."] = ".$running_str." ";
	}	
	else
	{
		for(my $i=1; $i<$n; $i++)
		{
			$rs .= $columns->[$i]." ";
		}
	}
	$rs;
}

# get_default_id($dbh,"nodeID","name","TLL_Nodelist",$cluster_id)
sub get_default_id
{
	my $dbh = $_[0];
	my $id1 = $_[1];
	my $name1 = $_[2];
	my $table = $_[3];
	my $primary_key_id = $_[4];

	my $default_id = undef;
	
	my $sql_str = "select ".$id1." from ".$table." where clusterID=".$primary_key_id." and (".$name1."='default' or ".$name1."='Default')";
	my $sth=$dbh->prepare($sql_str);
	my($id);
	if(!defined($sth->execute()))
	{
		error_and_exit($dbh->errstr);
	}
	else
	{
		$sth->bind_columns(undef,\$id);
		if($sth->fetch())
		{
			$default_id = $id;
		}
	}
	$default_id;		
}

# get_default_cluster_id($dbh,"clusterID","TLL_Cluster","name")
sub get_default_cluster_id
{
	my $dbh = $_[0];
	my $id1 = $_[1];
	my $table = $_[2];
	my $name1 = $_[3];

	my $default_id = undef;

	my $sql_str = "select ".$id1." from ".$table." where ".$name1."='default' or ".$name1."='Default'";
	my $sth=$dbh->prepare($sql_str);
	if(!defined($sth->execute()))
	{
		error_and_exit($dbh->errstr);
	}
	else
	{
		my($id);
		$sth->bind_columns(undef,\$id);
		if($sth->fetch())
		{
			$default_id = $id;
		}
	}
	$default_id;	
}

# get_default_id_by_primary_key($dbh,"classID","clusterID","TLLS_CFGClass","name",$classID)
sub get_default_id_by_primary_key
{
	my $dbh = $_[0];
	my $id1 = $_[1];
	my $id2 = $_[2];
	my $table = $_[3];
	my $name1 = $_[4];
	my $id_value = $_[5];

	my $default_id = undef;
	
	my $sql_str = "select ".$id2." from ".$table." where ".$id1."=".$id_value;
	my $sth=$dbh->prepare($sql_str);
	if(!defined($sth->execute()))
	{
		error_and_exit($dbh->errstr);
	}
	else
	{
		my($id);
		$sth->bind_columns(undef,\$id);
		if($sth->fetch())
		{
			$sql_str = "select ".$id1." from ".$table." where ".$id2."=".$id." and (".$name1."='default' or ".$name1."='Default')";
#			debug_log($sql_str);
			my $sth=$dbh->prepare($sql_str);
			if(!defined($sth->execute()))
			{
				error_and_exit($dbh->errstr);
			}
			else
			{
				$sth->bind_columns(undef,\$id);
				if($sth->fetch())
				{
					$default_id = $id;
				}
			}
		}
	}
	$default_id;	
}

# get_default_keyword_str($dbh,$sql_str,$table)
sub get_default_keyword_str
{
	my $dbh = $_[0];
	my $sql_str = $_[1];
	my $table = $_[2];

	my $default_str = "";
	
	my $sth = $dbh->prepare($sql_str);
	if(!defined($sth->execute()))
	{
		error_and_exit($dbh->errstr);
	}
	else
	{
		my $ary_ref;
		while($ary_ref = $sth->fetchrow_arrayref)
		{
			$default_str .= format_keywords($dbh,$ary_ref, $table)." ";
		}
	}
	debug_log("formatted keywords: ".$default_str);
	$default_str;
}

# get_default_items($dbh,$sql_str,$table)
sub get_default_items
{
	my $dbh = $_[0];
	my $sql_str = $_[1];
	my $table = $_[2];

#	debug_log($sql_str);
	
	my @default_items = ();

	my $sth = $dbh->prepare($sql_str);
	if(!defined($sth->execute()))
	{
		error_and_exit($dbh->errstr);
	}
	else
	{
		# in case there is no record found.
		for(my $m=1;$m<$sth->{NUM_OF_FIELDS};$m++)
		{
			my %item = (name => $sth->{NAME}->[$m],
				    value => "");
			push @default_items, \%item;
		}
		my $ary_ref = $sth->fetchrow_arrayref;
		if(defined($ary_ref))
		{
			@default_items = ();
			for(my $m=1;$m<$sth->{NUM_OF_FIELDS};$m++)
			{
				if(defined($ary_ref->[$m]))
				{
					my %item = (name => $sth->{NAME}->[$m],
						    value => $ary_ref->[$m]);
					push @default_items, \%item;
					
				}
				else
				{
					my %item = (name => $sth->{NAME}->[$m],
						    value => "");
					push @default_items, \%item;
				}
			}
		}
	}
	\@default_items;
}

# default_value_filter($table, $i_name)
# Filter out the hidden fields for the tables, no defaults needed for those
sub default_value_filter
{
	my $table = $_[0];
	my $name = lc($_[1]);

	my $ret = 0;
	
	if($table eq "TLLS_CFGClass")
	{
		if(lc($name) eq lc("bgListType") || lc($name) eq lc("groupListType") || lc($name) eq lc("userListType") || lc($name) eq lc("clusterID"))
		{
			$ret = -1;
		}
	}
	elsif($table eq "TLLS_CFGGroup")
	{
		if($name eq lc("clusterID") || $name eq lc("userListType"))
		{
			$ret = -1;
		}
	}
	elsif($table eq "TLLR_CFGRegion" || $table eq "TLLS_CFGUser" || $table eq "TLL_MachineGroup")
	{
		if($name eq lc("clusterID"))
		{
			$ret = -1;
		}
	}
	elsif($table eq "TLLR_CFGMCluster")
	{
		if($name eq lc("classListType") || $name eq lc("groupListType") || $name eq lc("userListType"))
		{
			$ret = -1;
		}
	}
	elsif($table eq "TLL_Cluster")
	{
		if($name eq lc("lastCFGMod") || $name eq lc("lldbVersion"))
		{
			$ret = -1;
		}
		
	}
	
	$ret;
}

# this function can not be applied to:
#                                     tab_panel_machine_env
#                                     tab_panel_class_rules
# get_default_json_str($dbh,"tab_panel_classes","classID",$default_class_id,$is_default)
sub get_default_json_str
{
	my $dbh = $_[0];
	my $panel = $_[1];
	my $primary_key_name = $_[2];
	my $primary_key_value = $_[3];
	my $is_default = $_[4];
	
	my $json_str = "";

	my $sql_str;

	my $n = scalar(@{LLUtil::ALL_TABLES->{$panel}});
	if($n == 0)
	{
		die;
	}
	
	for(my $i = 0; $i < $n; $i++)
	{
		my $type = LLUtil::ALL_TABLES->{$panel."_type"}->[$i];
		my @tables = @{LLUtil::ALL_TABLES->{$panel}->[$i]};
		if($type eq "normal")
		{
			foreach my $table (@tables)
			{
				$sql_str = "select * from ".$table." where ".$primary_key_name."=".$primary_key_value;
				my $rs = get_default_items($dbh,$sql_str,$table);
				my $n = scalar(@{$rs});
				for(my $i=0; $i<$n; $i++)
				{
					my $i_name = $rs->[$i]->{"name"};
					if(default_value_filter($table, $i_name) == 0)
					{
						if($is_default == 1)
						{
							# default item is selected in pull down
							$json_str .= "{type:'normal_default',id:'".$table."-".$i_name."-default',value:''},";
						}
						else
						{
							$json_str .= "{type:'normal_default',id:'".$table."-".$i_name."-default',value:'".$rs->[$i]->{"value"}."'},";
						}
					}
				}
			}
		}
		elsif($type eq "editable" || $type eq "multiselect")
		{
			foreach my $table (@tables)
			{
				my $default_str = "";
				$sql_str = "select * from ".$table." where ".$primary_key_name."=".$primary_key_value;
				$default_str = get_default_keyword_str($dbh,$sql_str,$table);
				if($is_default == 1)
				{
					# default item is selected in pull down
					$json_str .= "{type:'default_string',id:'".$table."_default_str"."',value:''},";
				}
				else
				{
					$json_str .= "{type:'default_string',id:'".$table."_default_str"."',value:'".$default_str."'},";
				}
			}
		}
		else
		{
			debug_log("Unknown table type: ".$type);
			die;
		}
		
	}
	
	$json_str;
}

# check_data($dbh,$table[1],$left_keys[1],$tmp[1],$return_name,"normal_add")
sub check_data
{
	my $dbh = $_[0];
	my $table = $_[1];
	my $column = $_[2];
	my $value = $_[3];
	my $name = $_[4];
	my $operation = $_[5];

	my $tmp_str = "table: ".$table."; column: ".$column."; value: ".$value."; name: ".$name;
	debug_log("check_data() ".$tmp_str);
	
	my $sql_str;
	my $keyword_exist = 0;
	my($keyword,$stanza);
	if (($table ne "TLLS_CFGClassLimits") && ($table ne "TLLR_CFGFileSysMonThresholds"))
	{
		if(defined($column))
		{
			$sql_str = "select keyword,stanza from TLL_CFGKeywords where tableName='".$table."' and columnName='".$column."'";
		}
		else
		{
			$sql_str = "select keyword,stanza from TLL_CFGKeywords where tableName='".$table."'";
		}
		debug_log("check_data() ".$sql_str);
		my $sth=$dbh->prepare($sql_str);
		if(!defined($sth->execute()))
		{
			error_and_exit($dbh->errstr);
		}
		$sth->bind_columns(undef,\$keyword,\$stanza);
		while($sth->fetch())
		{
			$keyword_exist = 1;
			debug_log("check_data(): keyword-".$keyword.", stanza-".$stanza);
		}
	}
	else
	{
		# for keyword from TLLS_CFGClassLimits or TLLR_CFGFileSysMonThresholds,
		# only check the input key-value pair,
		# and the keyword name is from input parameter
		$keyword_exist = 1;
		$keyword = $column;
		if ($table eq "TLLS_CFGClassLimits")
		{
			$stanza = "class";
		}
	}
	my $ret = 0;
	my $output = "";
	if($keyword_exist == 1)
	{
		my $cmd = "/var/www/ll/cgi-bin/llcheck.cgi";
		my $command_line = $cmd;
		if(!defined($stanza) || length($stanza) == 0)
		{
			$command_line .= " $keyword='$value' 2>&1";
		}
		else
		{
			if($stanza eq "class;user")
			{
				$stanza = "class";
			}
			$command_line .= " -t $stanza=$name $keyword='$value' 2>&1";
		}
		debug_log($command_line);
		$output = `$command_line`;
		$ret = $?;
# 		debug_log("Output: ".$ret);
# 		debug_log("Return code: ".$?);
	}
	else
	{
		debug_log("No record in TLL_CFGKeywords for table ".$table." and column ".$column);
	}
	if($ret != 0)
	{
		# return_key contains information for javascript to mark the invalid input.
		my $return_key = $operation."=".$table;
		if($operation eq "normal_add" || $operation eq "normal_update")
		{
			$return_key .= "-".$column;
		}
		
		print OUTPUT_HEADER."Check:".$return_key."&".$output;
		exit 1;
	}
}

#########################################################
#                                                       #
#               3. Functions for special panel          #
#                                                       #
#########################################################

# get_fsID_by_nodeID($dbh,$nodeID)
sub get_fsID_by_nodeID
{
	my $dbh = $_[0];
	my $nodeID = $_[1];
	
	my $sql_str="select fsID from TLLR_CFGFileSysMon where nodeID=".$nodeID;
	my $sth=$dbh->prepare($sql_str);
	if(!defined($sth->execute()))
	{
		error_and_exit($dbh->errstr);
	}
	my($fsID);
	$sth->bind_columns(undef,\$fsID);
	if(!$sth->fetch())
	{
		#todo: how to handle error here?
	}
	if(!defined($fsID))
	{
		$sql_str="insert into TLLR_CFGFileSysMon (nodeID) values(".$nodeID.")";
		if(!defined($dbh->do($sql_str)))
		{
			error_and_exit($dbh->errstr);
		}
		$fsID = $dbh->last_insert_id(undef,undef,undef,undef);
	}
	$fsID;
}

# get_default_json_str_of_machine_env($dbh,$default_node_id,$is_default)
sub get_default_json_str_of_machine_env
{
	my $dbh = $_[0];
	my $default_node_id = $_[1];
	my $is_default = $_[2];
	
	my @normal_tables_by_nodeID = @{LLUtil::ALL_TABLES->{"tab_panel_machine_env"}->[0]};
	my @editable_tables_by_nodeID = @{LLUtil::ALL_TABLES->{"tab_panel_machine_env"}->[1]};
	my @normal_tables_by_fsID = @{LLUtil::ALL_TABLES->{"tab_panel_machine_env"}->[2]};
	my @editable_tables_by_fsID = @{LLUtil::ALL_TABLES->{"tab_panel_machine_env"}->[3]};

	my $default_fs_id = get_fsID_by_nodeID($dbh,$default_node_id);

	my $sql_str;
	my $json_str = "";
	
	foreach my $table (@normal_tables_by_nodeID)
	{
		$sql_str = "select * from ".$table." where nodeID=".$default_node_id;
		my $rs = get_default_items($dbh,$sql_str,$table);
		my $n = scalar(@{$rs});
		for(my $i=0; $i<$n; $i++)
		{
			my $i_name = $rs->[$i]->{"name"};
			if(default_value_filter($table, $i_name) == 0)
			{
				if($is_default == 1)
				{
					# default item is selected in pull down
					$json_str .= "{type:'normal_default',id:'".$table."-".$i_name."-default',value:''},";
				}
				else
				{
					$json_str .= "{type:'normal_default',id:'".$table."-".$i_name."-default',value:'".$rs->[$i]->{"value"}."'},";
				}
			}
		}
	}

	foreach my $table (@editable_tables_by_nodeID)
	{
		my $default_str = "";
		$sql_str = "select * from ".$table." where nodeID=".$default_node_id;
		$default_str = get_default_keyword_str($dbh,$sql_str,$table);
		if($is_default == 1)
		{
			# default item is selected in pull down
			$json_str .= "{type:'default_string',id:'".$table."_default_str"."',value:''},";
		}
		else
		{
			$json_str .= "{type:'default_string',id:'".$table."_default_str"."',value:'".$default_str."'},";
		}
	}

	if(defined($default_fs_id))
	{
		foreach my $table (@normal_tables_by_fsID)
		{
			$sql_str = "select * from ".$table." where fsID=".$default_fs_id;
			my $rs = get_default_items($dbh,$sql_str,$table);
			my $n = scalar(@{$rs});
			for(my $i=0; $i<$n; $i++)
			{
				my $i_name = $rs->[$i]->{"name"};
				if(default_value_filter($table, $i_name) == 0)
				{
					if($is_default == 1)
					{
						# default item is selected in pull down
						$json_str .= "{type:'normal_default',id:'".$table."-".$i_name."-default',value:''},";
					}
					else
					{
						$json_str .= "{type:'normal_default',id:'".$table."-".$i_name."-default',value:'".$rs->[$i]->{"value"}."'},";
					}
				}
			}
		}
		
		foreach my $table (@editable_tables_by_fsID)
		{
			my $default_str = "";
			$sql_str = "select * from ".$table." where fsID=".$default_fs_id;
			$default_str = get_default_keyword_str($dbh,$sql_str,$table);
			if($is_default == 1)
			{
				# default item is selected in pull down
				$json_str .= "{type:'default_string',id:'".$table."_default_str"."',value:''},";
			}
			else
			{
				$json_str .= "{type:'default_string',id:'".$table."_default_str"."',value:'".$default_str."'},";
			}
		}
	}
	$json_str;
}

# get_outgoing_classes($dbh,$ary_ref->[0])
sub get_outgoing_classes
{
	my $dbh = $_[0];
	my $preemptID = $_[1];

	my $sql_str = "select * from TLLS_CFGPreemptClassOutgoing where preemptID=".$preemptID;

	my @ret = ();
	my $sth = $dbh->prepare($sql_str);
	if(!defined($sth->execute()))
	{
		error_and_exit($dbh->errstr);
	}
	else
	{
		my $ary_ref;
		while($ary_ref = $sth->fetchrow_arrayref)
		{
			push(@ret,$ary_ref->[1]);
		}
	}
	\@ret;
}

# get_running_tasks($dbh,$ary_ref->[0])
sub get_running_tasks
{
	my $dbh = $_[0];
	my $startID = $_[1];

	my $sql_str = "select * from TLLS_CFGPreemptStartClassRunningTasks where startID=".$startID;

	my @ret = ();
	my $sth = $dbh->prepare($sql_str);
	if(!defined($sth->execute()))
	{
		error_and_exit($dbh->errstr);
	}
	else
	{
		my $ary_ref;
		while($ary_ref = $sth->fetchrow_arrayref)
		{
			push(@ret,"(".$ary_ref->[1]."<".$ary_ref->[2].")");
		}
	}
	\@ret;	
}

# get_default_json_str_of_preempt($dbh,$default_cluster_id,$is_default)
sub get_default_json_str_of_preempt
{
	my $dbh = $_[0];
	my $default_cluster_id = $_[1];
	my $is_default = $_[2];
	
	# get default string for TLLS_CFGPreemptClass
	my $sql_str = "select * from TLLS_CFGPreemptClass where clusterID=".$default_cluster_id;

	my $default_str = "{type:'default_string',id:'TLLS_CFGPreemptClass_default_str',value:'";

	if($is_default == 0)
	{
		
		my $sth = $dbh->prepare($sql_str);
		if(!defined($sth->execute()))
		{
			error_and_exit($dbh->errstr);
		}
		else
		{
			my $ary_ref;
			while($ary_ref = $sth->fetchrow_arrayref)
			{
				$default_str .= format_keywords($dbh,$ary_ref,"TLLS_CFGPreemptClass");
			}
		}
		$default_str =~ s/ $//g;
	}
	
	$default_str .= "'},";

	# get default string for TLLS_CFGPreemptStartClass
	$sql_str = "select * from TLLS_CFGPreemptStartClass where clusterID=".$default_cluster_id;

	$default_str .= "{type:'default_string',id:'TLLS_CFGPreemptStartClass_default_str',value:'";
	if($is_default == 0)
	{
		my $sth = $dbh->prepare($sql_str);
		if(!defined($sth->execute()))
		{
			error_and_exit($dbh->errstr);
		}
		else
		{
			my $ary_ref;
			while($ary_ref = $sth->fetchrow_arrayref)
			{
				$default_str .= format_keywords($dbh,$ary_ref,"TLLS_CFGPreemptStartClass");
			}
		}
		$default_str =~ s/ $//g;
	}
	
	$default_str .= "'},";
	
	$default_str;	
}


# get_admin()
sub get_admin
{
        my ($db_host, $db_user, $db_pass, $db_db, $db_conn, $db2inst) = get_db_info(0);
        # get the first id in the admin list that exists on the node
        my $json_str = "Content-type: text/html\n\n";
        my $dbh = get_dbh(0);
        if(!defined($dbh))
        {
               $json_str .= DBI->errstr;
               print $json_str;
               exit 1;
        }

        # Get the list of admin names from db.
        my $admin_name;
        my $sql_str="select admin_list from TLL_CFGCluster".get_only_one($db_conn);
        my $sth=$dbh->prepare($sql_str);
        if(defined($sth->execute()))
        {
                my($admin);
                $sth->bind_columns(undef,\$admin);
                while($sth->fetch())
                {
                        my @tmp = split(' ',$admin);
                        debug_log("get_admin: testing admin_list:".$admin);

                        # test to ensure that admin id exists on this machine
                        foreach my $test_admin (@tmp)
                        {
                                debug_log("get_admin: testing an admin:".$test_admin);
                                my $garbage = `/usr/bin/id $test_admin`;
                                # found an admin id
                                if($? == 0)
                                {
                                	debug_log("get_admin: found good admin:".$test_admin);
                                        $admin_name=$test_admin;
                                        last;
                                }
                        }
                        if (length($admin_name) == 0)
                        {
                                # no name from admin_list is on this machine
                                debug_log("get_admin: None of the LoadLeveler admin IDs exist on the node where reconfig will be issued.");
                        }

                }
        }
        else
        {
                $json_str .= $dbh->errstr;
                print $json_str;
                exit 1;
        }
	$sth->finish();
	$dbh->disconnect();
	$admin_name;
}

1;
